@@ -20,20 +20,26 @@ from django.utils.translation import ugettext_lazy as _
 from django.utils.encoding import force_unicode
 from django.conf import settings
 from django import http
+from django.db import models
+#~ from django.core.mail import EmailMultiAlternatives
+from django.core.mail import EmailMessage
 
 
 import lino
 from lino.utils import AttrDict
 from lino.utils import curry
 from lino.utils import babel
-from lino.utils import Warning
+#~ from lino.utils import Warning
 
 from lino.ui import requests as ext_requests
 
-from lino.core.modeltools import resolve_model
+from lino.core.modeltools import resolve_model, resolve_app
+from lino.core import layouts
+#~ from lino.core import changes
+from lino.core import fields
 
 #~ from lino.core.perms import UserLevels
-#~ from lino.core import perms 
+from lino.core import perms 
 
 
 class VirtualRow(object):
@@ -44,8 +50,8 @@ class VirtualRow(object):
         for k,v in kw.items():
             setattr(self,k,v)
 
-    def get_row_permission(self,user,state,action):
-        if action.readonly:
+    def get_row_permission(self,ar,state,ba):
+        if ba.action.readonly:
             return True
         return False 
             
@@ -120,13 +126,121 @@ class ConfirmationRequired(Exception):
         self.step = step
         self.messages = messages
         Exception.__init__(self)
+        
+class DialogRequired(Exception):
+    """
+    This is the special exception risen when an Action calls 
+    :meth:`ActionRequest.dialog`.
+    """
+    def __init__(self,step,dlg):
+        self.step = step
+        self.dialog = dlg
+        Exception.__init__(self)
+        
 
 
-class Action(object):
+#~ @classmethod
+def register_params(cls):
+    if cls.parameters:
+        for k,v in cls.parameters.items():
+            v.set_attributes_from_name(k)
+            v.table = cls
+        if cls.params_layout is None:
+            cls.params_layout = cls._layout_class.join_str.join(cls.parameters.keys())
+        if isinstance(cls.params_layout,basestring):
+            cls.params_layout = cls._layout_class(cls.params_layout,cls)
+        elif isinstance(cls.params_layout,layouts.Panel):
+            cls.params_layout = cls._layout_class(cls.params_layout.desc,cls,**cls.params_layout.options)
+
+def setup_params_choosers(self):
+    if self.parameters:
+        from lino.utils.choosers import check_for_chooser
+        for k,fld in self.parameters.items():
+            if isinstance(fld,models.ForeignKey):
+                fld.rel.to = resolve_model(fld.rel.to)
+            check_for_chooser(self,fld)
+
+def make_params_layout_handle(self,ui):
+    return self.params_layout.get_layout_handle(ui)
+        
+class Parametrizable(object):        
+  
+    active_fields = None # 20121006
+    master_field = None
+    
+    parameters = None
+    """
+    User-definable parameter fields for this table.
+    Set this to a `dict` of `name = models.XyzField()` pairs.
+    """
+    
+    #~ params_template = None # no longer used
+    
+    params_layout = None
+    """
+    If this table has parameters, specify here how they should be 
+    laid out in the parameters panel.
+    """
+    
+    params_panel_hidden = False
+    """
+    If this table has parameters, set this to True if the parameters 
+    panel should be initially hidden when this table is being displayed.
+    """
+    
+    _layout_class = NotImplementedError
+    
+            
+    #~ @classmethod
+    #~ def install_params_on_actor(cls):
+        #~ for k in ('get_window_layout', 'after_site_setup',
+          #~ 'make_params_layout_handle','get_param_elem'):
+            #~ um = getattr(cls,k)
+            #~ setattr(cls,k,classmethod(um))
+        
+        
+    @classmethod
+    def get_param_elem(self,name):
+        if self.parameters:
+            return self.parameters.get(name,None)
+        #~ for pf in self.params:
+            #~ if pf.name == name:  return pf
+        return None
+      
+    #~ @classmethod
+    def get_window_layout(self,actor):
+        return self.params_layout
+        
+        
+#~ class ActionMetaClass(type):
+    #~ def __new__(meta, classname, bases, classDict):
+        #~ cls = type.__new__(meta, classname, bases, classDict)
+        #~ cls.register_params()
+        #~ return cls
+      
+
+
+class Action(Parametrizable):
     """
     Abstract base class for all Actions
     """
     
+    #~ __metaclass__ = ActionMetaClass
+    
+    debug_permissions = False
+    
+    icon_name = None 
+    """
+    The class name of an icon to be used for this action.
+    """
+    
+    icon_file = None
+    """
+    The file name of an icon to be used for this action.
+    """
+    
+    _layout_class = layouts.ActionParamsLayout
+    
     sort_index = 90
     """
     
@@ -135,10 +249,12 @@ class Action(object):
     ===== =================================
     value action
     ===== =================================
-    20    :class:`Insert <InsertRow>`
-    30    :class:`Delete <DeleteSelected>`
+    10    :class:`insert <InsertRow>`
+    11    :attr:`duplicate <lino.mixins.duplicable.Duplicable.duplicate_row>`
+    20    :class:`detail <ShowDetailAction>`
+    30    :class:`delete <DeleteSelected>`
     50    :class:`Print <lino.mixins.printable.BasePrintAction>`
-    60    :attr:`Duplicate <lino.mixins.duplicable.Duplicable.duplicate_row>`
+    51    :class:`Clear Cache <lino.mixins.printable.ClearCacheAction>`
     90    default for all custom row actions created using :func:`@dd.action <action>`
     ===== =================================
     
@@ -170,19 +286,26 @@ class Action(object):
     See :func:`lino.core.perms.make_permission_handler`.
     """
     
-    name = None
+    action_name = None
     """
-    Internally used to store the name of this action within the Actor's namespace.
+    Internally used to store the name of this action within the defining Actor's namespace.
     """
     
-    url_action_name = None
+    #~ url_action_name = None
+    #~ """
+    #~ """
+    
+    use_param_panel = False
     """
+    Used internally. This is True for window actions whose window 
+    use the parameter panel: grid and emptytable (but not showdetail)
     """
     
-    actor = None
+    
+    defining_actor = None
     """
     Internally used to store the :class:`lino.core.actors.Actor` 
-    who owns this action.
+    who defined this action.
     """
     
     key = None
@@ -206,7 +329,18 @@ class Action(object):
     readonly = True
     """
     Whether this action possibly modifies data *in the given object*.
+    
     This means that :class:`InsertRow` is a `readonly` action.
+    Actions like :class:`InsertRow` and :class:`Duplicable <lino.mixins.duplicable.Duplicate>` 
+    which do not modify the given object but *do* modify the database,
+    must override their `get_action_permission`::
+    
+      def get_action_permission(self,ar,obj,state):
+          if user.profile.readonly: 
+              return False
+          return super(Duplicate,self).get_action_permission(ar,obj,state)
+        
+    
     """
     
     opens_a_window = False
@@ -216,7 +350,8 @@ class Action(object):
     
     hide_top_toolbar = False
     """
-    Used internally if :attr:`opens_a_window` to say whether the window has a top toolbar.
+    Used internally if :attr:`opens_a_window` to say whether 
+    the window has a top toolbar.
     """
     
     hide_navigator = False
@@ -234,7 +369,7 @@ class Action(object):
     """
     Used internally.
     Whether this action should be displayed 
-    as the :meth:`workflow_buttons <lino.core.actors.Actor.workflow_buttons>`.
+    as the :meth:`workflow_buttons <lino.core.model.Model.workflow_buttons>`.
     """
     
     custom_handler = False
@@ -258,6 +393,9 @@ class Action(object):
         if label is not None:
             self.label = label
             
+        register_params(self)
+        #~ setup_params_choosers(self.__class__)
+            
         #~ if label is None:
             #~ label = self.label or self.url_action_name 
         for k,v in kw.items():
@@ -268,7 +406,10 @@ class Action(object):
         #~ self.set_required(**required)
         assert self.callable_from is None or isinstance(
             self.callable_from,(tuple,type)), "%s" % self
+            
 
+    def make_params_layout_handle(self,ui):
+        return make_params_layout_handle(self,ui)
         
     def set_required(self,**kw):
         """
@@ -280,41 +421,71 @@ class Action(object):
         new.update(self.required)
         new.update(kw)
         self.required = new
-        #~ if isinstance(self,StateAction):
-        if self.required.has_key('states'):
-            self.show_in_workflow = True
+        #~ if self.required.has_key('states'):
+            #~ self.show_in_workflow = True
+            #~ self.custom_handler = True
+            #~ self.show_in_bbar = False
+        #~ else:
+            #~ self.show_in_workflow = False
+            #~ self.show_in_bbar = True
+        if self.show_in_workflow:
             self.custom_handler = True
             self.show_in_bbar = False
         else:
-            self.show_in_workflow = False
             self.show_in_bbar = True
         
+    def full_name(self,actor):
+        if self.action_name is None:
+            raise Exception("Tried to full_name() on %r" % self)
+            #~ return repr(self)
+        if self.parameters:
+            return self.defining_actor.actor_id + '.' + self.action_name
+        return str(actor) + '.' + self.action_name
+        
+    def __repr__(self):
+        return "%s %s.%s" % (self.__class__.__name__,self.defining_actor,self.action_name)
+        
     def __str__(self):
-        #~ raise Exception("Must use action2str(actor,action)")
-        if self.actor is None:
-            #~ raise Exception("tried to call str() on general action %s" % self.name)
+        if self.defining_actor is None:
             return repr(self)
-            #~ raise Exception("Tried to call str() on shared action %r" % self)
-        if self.name is None:
+        return unicode(self.defining_actor.label).encode('ascii','replace') + ' : ' + unicode(self.label).encode('ascii','replace')
+        
+    def unused__str__(self):
+        raise Exception("20121003 Must use full_name(actor)") 
+        if self.defining_actor is None:
             return repr(self)
-        return str(self.actor) + '.' + self.name
+        if self.action_name is None:
+            return repr(self)
+        return str(self.defining_actor) + ':' + self.action_name
         
     #~ def set_permissions(self,*args,**kw)
         #~ self.permission = perms.factory(*args,**kw)
         
+    def attach_to_workflow(self,wf,name):
+        assert self.action_name is None
+        self.action_name = name
+        self.defining_actor = wf
+        #~ logger.info("20121009 %r attach_to_workflow(%s)",self,self.full_name(wf))
+        
     def attach_to_actor(self,actor,name):
-        if self.name is not None:
-            raise Exception("%s tried to attach named action %s" % (actor,self))
-        if self.actor is not None:
-            raise Exception("%s tried to attach action %s of %s" % (actor,name,self.actor))
-        self.name = name
-        self.actor = actor
+        #~ if self.name is not None:
+            #~ raise Exception("%s tried to attach named action %s" % (actor,self))
+        #~ if actor == self.defining_actor:
+            #~ raise Exception('20121003 %s %s' % (actor,name))
+        if self.defining_actor is not None:
+            # already defined in another actor
+            return
+        if self.action_name is not None:
+            raise Exception("tried to attach named action %s.%s" % (actor,self.action_name))
+        self.action_name = name
+        self.defining_actor = actor
         if self.label is None:
             self.label = name
-        if actor.hide_top_toolbar:
-            self.hide_top_toolbar = True
-        if self.help_text is None and self is actor.default_action:
-            self.help_text  = actor.help_text
+        #~ if actor.hide_top_toolbar:
+            #~ self.hide_top_toolbar = True
+        #~ if self.help_text is None \
+            #~ and name == actor.get_default_action():
+            #~ self.help_text  = actor.help_text
         #~ if name == 'default_action':
             #~ print 20120527, self
             
@@ -337,26 +508,29 @@ class Action(object):
         #~ """
         #~ return True
         
-    def get_button_label(self):
-        if self.actor is None:
-            return self.label 
-        if self is self.actor.default_action:
-            return self.label 
-        else:
-            return u"%s %s" % (self.label,self.actor.label)
-            
-    def get_action_permission(self,user,obj,state):
+    def get_action_permission(self,ar,obj,state):
         """
-        The default implementation simply calls this action's 
-        permission handler.
         Derived Action classes may override this to add vetos.
         E.g. DispatchAction is not available for a User with empty partner.
         """
-        return self.allow(user,obj,state)
+        return True
         
     #~ def run(self,elem,ar,**kw):
         #~ raise NotImplementedError("%s has no run() method" % self.__class__)
 
+    #~ def request(self,*args,**kw):
+        #~ kw.update(action=self)
+        #~ return self.defining_actor.request(*args,**kw)
+        
+    def action_param_defaults(self,ar,obj,**kw):
+        """
+        Same as Actor.param_defaults, except that here it is a instance method
+        """
+        for k,pf in self.parameters.items():
+            kw[k] = pf.get_default()
+        return kw
+        
+        
 
 class TableAction(Action):
     """
@@ -386,47 +560,14 @@ class RowAction(Action):
         raise NotImplementedError("%s has no run() method" % self.__class__)
 
     #~ def get_action_permission(self,user,obj):
-        #~ return self.actor.get_row_permission(self,user,obj)
+        #~ return self.actor.get_row_permission(self,ar,obj)
             
-    def attach_to_actor(self,actor,name):
-        super(RowAction,self).attach_to_actor(actor,name)
-        if not self.url_action_name:
-            self.url_action_name = name 
+    #~ def attach_to_actor(self,actor,name):
+        #~ super(RowAction,self).attach_to_actor(actor,name)
+        #~ if not self.url_action_name:
+            #~ self.url_action_name = name 
 
 
-class StateAction(RowAction):
-    ajax = True
-    def __init__(self,actor,target_state,**kw):
-        self.target_state = target_state
-        kw.update(label=getattr(target_state,'action_label',target_state.text))
-        required = getattr(target_state,'required',None)
-        if required is not None:
-            if target_state.name:
-                m = getattr(actor.model,'allow_state_'+target_state.name,None)
-                if m is not None:
-                    def allow(action,user,obj,state):
-                        return m(obj,user)
-                    required.update(allow=allow)
-            kw.update(required=required)
-        help_text = getattr(target_state,'help_text',None)
-        if help_text:
-            kw.update(help_text=help_text)
-        super(StateAction,self).__init__(**kw)
-        #~ print 20120709, self, self.show_in_workflow
-        
-    def run(self,row,ar,**kw):
-        state_field_name = self.actor.workflow_state_field.attname
-        assert isinstance(state_field_name,basestring)
-        #~ old = row.state
-        old = getattr(row,state_field_name)
-        row.before_state_change(ar,kw,old,self.target_state)
-        #~ row.state = self.target_state
-        setattr(row,state_field_name,self.target_state)
-        row.save()
-        row.after_state_change(ar,kw,old,self.target_state)
-        return ar.ui.success_response(**kw)
-        
-    
 
 class RedirectAction(Action):
     
@@ -438,18 +579,20 @@ class RedirectAction(Action):
 
 class GridEdit(TableAction):
   
+    use_param_panel = True
     show_in_workflow = False
     opens_a_window = True
+    #~ icon_file = 'application_view_list.png' # used e.g. by quick_add_buttons()
 
     callable_from = tuple()
-    url_action_name = 'grid'
+    action_name = 'grid'
     
     def attach_to_actor(self,actor,name):
         #~ self.label = actor.button_label or actor.label
         self.label = actor.label
         super(GridEdit,self).attach_to_actor(actor,name)
 
-    def get_window_layout(self):
+    def get_window_layout(self,actor):
         #~ return self.actor.list_layout
         return None
 
@@ -458,18 +601,21 @@ class ShowDetailAction(RowAction):
     """
     An action that opens the Detail Window of its actor.
     """
+    icon_name = 'x-tbar-detail'
+    #~ icon_file = 'application_form.png'
     opens_a_window = True
     show_in_workflow = False
     
-    #~ sort_index = 1
+    sort_index = 20
     callable_from = (GridEdit,)
     #~ show_in_detail = False
     #~ needs_selection = True
-    url_action_name = 'detail'
+    action_name = 'detail'
     label = _("Detail")
+    help_text = _("Open a detail window on this record")
     
-    def get_window_layout(self):
-        return self.actor.detail_layout
+    def get_window_layout(self,actor):
+        return actor.detail_layout
         
     #~ def get_elem_title(self,elem):
         #~ return _("%s (Detail)")  % unicode(elem)
@@ -484,15 +630,18 @@ class InsertRow(TableAction):
     window gets submitted.
     """
     label = _("New")
+    icon_name = 'x-tbar-new' # if action rendered as toolbar button
+    icon_file = 'add.png' # if action rendered by quick_add_buttons
     show_in_workflow = False
     opens_a_window = True
     hide_navigator = True
-    sort_index = 20
+    sort_index = 10
     hide_top_toolbar = True
+    help_text = _("Insert a new record")
     #~ readonly = False # see blog/2012/0726
     required = dict(user_level='user')
     callable_from = (GridEdit,ShowDetailAction)
-    url_action_name = 'insert'
+    action_name = 'insert'
     #~ label = _("Insert")
     key = INSERT # (ctrl=True)
     #~ needs_selection = False
@@ -500,14 +649,14 @@ class InsertRow(TableAction):
     def get_action_title(self,rr):
         return _("Insert into %s") % force_unicode(rr.get_title())
         
-    def get_window_layout(self):
-        return self.actor.insert_layout or self.actor.detail_layout
+    def get_window_layout(self,actor):
+        return actor.insert_layout or actor.detail_layout
 
-    def get_action_permission(self,user,obj,state):
+    def get_action_permission(self,ar,obj,state):
         # see blog/2012/0726
-        if user.profile.readonly: 
+        if ar.get_user().profile.readonly: 
             return False
-        return super(InsertRow,self).get_action_permission(user,obj,state)
+        return super(InsertRow,self).get_action_permission(ar,obj,state)
 
 
 
@@ -518,16 +667,18 @@ class DuplicateRow(RowAction):
     readonly = False
     required = dict(user_level='user')
     callable_from = (GridEdit,ShowDetailAction)
-    url_action_name = 'duplicate'
+    action_name = 'duplicate'
     label = _("Duplicate")
 
 
 class ShowEmptyTable(ShowDetailAction):
+    use_param_panel = True
     callable_from = tuple()
-    url_action_name = 'show' 
+    action_name = 'show' 
     default_format = 'html'
     #~ hide_top_toolbar = True
     hide_navigator = True
+    icon_name = None
     
     def attach_to_actor(self,actor,name):
         self.label = actor.label
@@ -558,6 +709,8 @@ class DeleteSelected(RowAction):
     """
     Delete the row.
     """
+    icon_name = 'x-tbar-delete'
+    help_text = _("Delete this record")
     auto_save = False
     sort_index = 30
     readonly = False
@@ -566,12 +719,16 @@ class DeleteSelected(RowAction):
     callable_from = (GridEdit,ShowDetailAction)
     #~ needs_selection = True
     label = _("Delete")
-    url_action_name = 'delete'
+    #~ url_action_name = 'delete'
     key = DELETE # (ctrl=True)
     #~ client_side = True
     
         
 class SubmitDetail(RowAction):
+    sort_index = 10
+    switch_to_detail = False
+    icon_name = 'x-tbar-save'
+    help_text = _("Save changes in this form")
     label = _("Save")
     auto_save = False
     show_in_workflow = False
@@ -582,13 +739,142 @@ class SubmitDetail(RowAction):
     callable_from = (ShowDetailAction,)
     
 class SubmitInsert(SubmitDetail):
+    sort_index = 10
+    switch_to_detail = True
+    icon_name = None # don't inherit 'x-tbar-save' from Submitdetail 
     #~ url_action_name = 'SubmitInsert'
     label = _("Create")
+    help_text = _("Create the record and open a detail window on it")
     #~ label = _("Insert")
     callable_from = (InsertRow,)
+    
+class SubmitInsertAndStay(SubmitInsert):
+    sort_index = 11
+    switch_to_detail = False
+    label = _("Create without detail")
+    help_text = _("Don't open an detail window on the new record")
+
+    
+    
+class NotifyingAction(RowAction):
+    
+    parameters = dict(
+        notify_subject = models.CharField(_("Summary"),blank=True,max_length=200),
+        notify_body = fields.RichTextField(_("Description"),blank=True),
+        notify_silent = models.BooleanField(_("Don't send email notification"),default=False),
+    )
+    
+    params_layout = layouts.Panel("""
+    notify_subject
+    notify_body
+    notify_silent
+    """,window_size=(50,15))
+    
+    #~ def update_system_note_kw(self,ar,kw,obj):
+        #~ pass
+        
+    def get_notify_subject(self,ar,obj):
+        """
+        Return the default value of the `notify_subject` field.
+        """
+        return None
+        
+    def get_notify_body(self,ar,obj):
+        """
+        Return the default value of the `notify_body` field.
+        """
+        return None
+        
+    def action_param_defaults(self,ar,obj,**kw):
+        kw = super(NotifyingAction,self).action_param_defaults(ar,obj,**kw)
+        s = self.get_notify_subject(ar,obj)
+        if s is not None: kw.update(notify_subject=s)
+        s = self.get_notify_body(ar,obj)
+        if s is not None: kw.update(notify_body=s)
+        return kw
+        
+    def run(self,obj,ar,**kw):
+        kw.update(message=ar.action_param_values.notify_subject)
+        kw.update(alert=True)
+        kw = super(NotifyingAction,self).run(obj,ar,**kw)
+        self.add_system_note(ar,obj)
+        return kw
+    
+    def add_system_note(self,ar,owner,**kw):
+        #~ body = _("""%(user)s executed the following action:\n%(body)s
+        #~ """) % dict(user=ar.get_user(),body=body)
+        ar.add_system_note(
+            owner,
+            ar.action_param_values.notify_subject,
+            ar.action_param_values.notify_body,
+            ar.action_param_values.notify_silent,**kw)
+        
+    
 
 
-
+class BoundAction(object):
+  
+    def __init__(self,actor,action):
+        if not isinstance(action,Action):
+            raise Exception("%s : %r is not an Action" % (actor,action))
+        self.action = action
+        self.actor = actor
+        
+        
+        required = dict()
+        if action.readonly:
+            required.update(actor.required)
+        #~ elif isinstance(action,InsertRow):
+            #~ required.update(actor.create_required)
+        elif isinstance(action,DeleteSelected):
+            required.update(actor.delete_required)
+        else:
+            required.update(actor.update_required)
+        required.update(action.required)
+        #~ print 20120628, str(a), required
+        #~ def wrap(a,required,fn):
+            #~ return fn
+            
+        #~ a.allow = curry(wrap(a,required,perms.make_permission_handler(
+            #~ a,actor,a.readonly,actor.debug_permissions,**required)),a)
+        #~ ba = actions.BoundAction(actor,a)
+        debug = actor.debug_permissions or action.debug_permissions
+        self.allow = curry(perms.make_permission_handler(
+            action,actor,action.readonly,debug,**required),action)
+        #~ actor.actions.define(a.action_name,ba)
+        
+        
+    def get_window_layout(self):
+        return self.action.get_window_layout(self.actor)
+        
+    def full_name(self):
+        return self.action.full_name(self.actor)
+        #~ if self.action.action_name is None:
+            #~ raise Exception("%r action_name is None" % self.action)
+        #~ return str(self.actor) + '.' + self.action.action_name
+        
+    def request(self,*args,**kw):
+        kw.update(action=self)
+        return self.actor.request(*args,**kw)
+        
+        
+    def get_button_label(self):
+        if self.actor is None:
+            return self.action.label 
+        if self.action is self.actor.default_action.action:
+            return self.actor.label 
+        else:
+            return u"%s %s" % (self.action.label,self.actor.label)
+            
+        
+    def get_bound_action_permission(self,ar,obj,state):
+        if not self.action.get_action_permission(ar,obj,state):
+            return False
+        return self.allow(ar.get_user(),obj,state)
+        
+    def get_view_permission(self,user):
+        return self.allow(user,None,None)
+        
 
 class ActionRequest(object):
     """
@@ -605,7 +891,18 @@ class ActionRequest(object):
     create_kw = None
     renderer = None
     
-    def __init__(self,ui,actor,request=None,action=None,renderer=None,**kw):
+    offset = None
+    limit = None
+    order_by = None
+    
+    def __init__(self,ui,actor,request=None,action=None,renderer=None,param_values=None,**kw):
+        """
+        An ActionRequest is instantiated from different shortcut methods:
+        
+        - :meth:`lino.core.actors.Actor.request`
+        - :meth:`lino.core.actions.Action.request`
+        
+        """
         #~ ActionRequest.__init__(self,ui,action)
         if ui is None:
             ui = settings.LINO.ui
@@ -613,11 +910,17 @@ class ActionRequest(object):
         self.ui = ui
         self.error_response = ui.error_response
         self.success_response = ui.success_response
+        if renderer is None:
+            renderer = ui.text_renderer
         self.renderer = renderer
-        self.action = action or actor.default_action
+        self.actor = actor
+        #~ self.action = action or actor.default_action
+        #~ self.bound_action = BoundAction(actor,action or actor.default_action)
+        if action and not isinstance(action,BoundAction):
+            raise Exception("20121003 %r is not a BoundAction" % action)
+        self.bound_action = action or actor.default_action
         self.step = 0 # confirmation counter
         #~ self.report = actor
-        self.actor = actor
         self.request = request
         if request is not None:
             if request.method == 'PUT':
@@ -627,25 +930,63 @@ class ActionRequest(object):
             kw = self.parse_req(request,rqdata,**kw)
         #~ 20120605 self.ah = actor.get_handle(ui)
         self.setup(**kw)
-        # 20120605 : Actor.get_request_handle() ar instead of ui
         self.ah = actor.get_request_handle(self)
-        
-        if self.actor.parameters and request is not None:
-            param_values = self.ui.parse_params(self.ah,request)
+        """
+        See 20120825
+        """
+        #~ if self.actor.parameters is None:
+            #~ if param_values is not None:
+                #~ raise Exception("Cannot request param_values on %s" % self.actor)
+        #~ else:
+        if self.actor.parameters is not None:
+            pv = self.actor.param_defaults(self)
+            
+            """
+            New since 20120913.
+            E.g. newcomers.Newcomers is a simple pcsw.Clients with known_values=dict(client_state=newcomer)
+            and since there is a parameter `client_state`, we override that parameter's default value.
+            """
+            for k,v in self.known_values.items():
+                if pv.has_key(k):
+                    pv[k] = v
+            """
+            New since 20120914.
+            MyClientsByGroup has a known group, this 
+            must also appear as `group` parameter value.
+            Lino now understands tables where the master_key is also a parameter.
+            """
+            if self.actor.master_key is not None:
+                if pv.has_key(self.actor.master_key):
+                    pv[self.actor.master_key] = self.master_instance
                 
-            for k,pf in self.actor.parameters.items():
-                v = param_values.get(k,None)
-                if v is None:
-                    v = pf.get_default()
-                self.param_values.define(k,v)
-            if param_values:
-                #~ logger.info("20120608 param_values is %s",param_values)
-                for k,v in param_values.items():
-                    self.param_values.define(k,v)
+            if param_values is None:
+              if request is not None: # 20121025
+                #~ pv.update(self.ui.parse_params(self.ah,request))
+                #~ pv.update(self.ah.store.parse_params(request))
+                pv.update(self.actor.params_layout.params_store.parse_params(request))
+                
+            if param_values is not None:
+                for k in param_values.keys(): 
+                    if not pv.has_key(k):
+                        raise Exception("Invalid key %r in param_values" % k)
+                pv.update(param_values)
+                
+            self.param_values = AttrDict(**pv)
+            
+        if self.bound_action.action.parameters is not None:
+            apv = self.bound_action.action.action_param_defaults(self,None)
+            if request is not None:
+                #~ pv.update(self.ui.parse_params(self.ah,request))
+                #~ pv.update(self.ah.store.parse_params(request))
+                apv.update(self.bound_action.action.params_layout.params_store.parse_params(request))
+                
+            self.action_param_values = AttrDict(**apv)
+            
+            #~ if param_values:
+                #~ # logger.info("20120608 param_values is %s",param_values)
+                #~ for k,v in param_values.items():
+                    #~ self.param_values.define(k,v)
                 
-        
-    def href_to(self,*args,**kw):
-        return self.renderer.href_to(self,*args,**kw)
         
     def parse_req(self,request,rqdata,**kw): 
         #~ if self.actor.parameters:
@@ -666,7 +1007,7 @@ class ActionRequest(object):
     def setup(self,
             user=None,
             subst_user=None,
-            param_values={},
+            #~ param_values={},
             known_values=None,
             requesting_panel=None,
             renderer=None,
@@ -675,9 +1016,8 @@ class ActionRequest(object):
         self.user = user
         if renderer is not None:
             self.renderer = renderer
-        if self.actor.parameters:
-            self.param_values = AttrDict(**param_values)
-            #~ self.param_values = param_values
+        #~ if self.actor.parameters:
+            #~ self.param_values = AttrDict(**param_values)
         self.subst_user = subst_user
         #~ 20120111 
         #~ self.known_values = known_values or self.report.known_values
@@ -690,6 +1030,13 @@ class ActionRequest(object):
         self.known_values = kw
         
         
+    def dialog(self,dlg):
+        # not finished
+        self.step += 1
+        if int(self.request.REQUEST.get(ext_requests.URL_PARAM_ACTION_STEP,'0')) >= self.step:
+            return
+        raise DialogRequired(self.step,dlg)
+        
     def confirm(self,*messages):
         """
         Calling this from an Action's :meth:`Action.run` method will
@@ -710,16 +1057,23 @@ class ActionRequest(object):
             return
         raise ConfirmationRequired(self.step,messages)
         
-    def create_phantom_row(self,**kw):
-        if self.create_kw is None or not self.actor.editable:
+    def create_phantom_rows(self,**kw):
+        #~ logger.info('20121011 %s.create_phantom_rows(), %r', self,self.create_kw)
+        if self.create_kw is None or not self.actor.editable or not self.actor.allow_create:
             #~ logger.info('20120519 %s.create_phantom_row(), %r', self,self.create_kw)
             return 
-        #~ if not self.actor.get_permission(self.get_user(),self.actor.create_action):
-        #~ if not self.actor.allow_create(self.get_user(),None,None):
-        if self.actor.create_action is not None:
-            if not self.actor.create_action.allow(self.get_user(),None,None):
-                return
-        return PhantomRow(self,**kw)
+        if not self.actor.get_create_permission(self):
+            return
+        #~ # if not self.actor.get_permission(self.get_user(),self.actor.create_action):
+        #~ # if not self.actor.allow_create(self.get_user(),None,None):
+        #~ # ca = self.actor.get_url_action('create_action')
+        #~ ca = self.actor.create_action
+        #~ if ca is not None:
+            #~ # if not self.actor.create_action.allow(self.get_user(),None,None):
+            #~ # if not ca.allow(self.get_user(),None,None):
+            #~ if not ca.get_bound_action_permission(self.get_user(),None,None):
+                #~ return
+        yield PhantomRow(self,**kw)
       
     def create_instance(self,**kw):
         if self.create_kw:
@@ -737,6 +1091,7 @@ class ActionRequest(object):
                 #~ kw[k] = v
         return obj
         
+        
     def get_data_iterator(self):
         raise NotImplementedError
         
@@ -755,24 +1110,38 @@ class ActionRequest(object):
         return self.subst_user or self.user
         
     def get_action_title(self):
-        return self.action.get_action_title(self)
+        return self.bound_action.action.get_action_title(self)
         
     def get_title(self):
         return self.actor.get_title(self)
         
     def render_to_dict(self):
-        return self.action.render_to_dict(self)
+        return self.bound_action.action.render_to_dict(self)
         
     def get_request_url(self,*args,**kw):
         return self.ui.get_request_url(self,*args,**kw)
 
+    def get_action_status(self,ba,obj,**kw):
+        #~ logger.info("get_action_status %s",ba.full_name())
+        if ba.action.parameters:
+            if ba.action.params_layout.params_store is None:
+                raise Exception("20121016 %s has no store" % ba.action.params_layout)
+            kw.update(field_values=ba.action.params_layout.params_store.pv2dict(
+                self.ui,ba.action.action_param_defaults(self,obj)))
+        return kw
+      
+      
     def get_status(self,ui,**kw):
         if self.actor.parameters:
-            kw.update(param_values=self.ah.store.pv2dict(ui,self.param_values))
+            #~ kw.update(param_values=self.ah.store.pv2dict(ui,self.param_values))
+            #~ lh = self.actor.params_layout.get_layout_handle(ui)
+            kw.update(param_values=self.actor.params_layout.params_store.pv2dict(ui,self.param_values))
         bp = kw.setdefault('base_params',{})
         if self.subst_user is not None:
             #~ bp[ext_requests.URL_PARAM_SUBST_USER] = self.subst_user.username
             bp[ext_requests.URL_PARAM_SUBST_USER] = self.subst_user.id
+        #~ if self.actor.__name__ == 'MyClients':
+            #~ print "20120918 actions.get_status", kw
         return kw
         
 
@@ -791,11 +1160,55 @@ class ActionRequest(object):
             actor = self.actor
         return self.ui.request(actor,**kw)
         
+    def href_to(self,*args,**kw): return self.renderer.href_to(self,*args,**kw)
+    def pk2url(self,*args,**kw): return self.renderer.pk2url(self,*args,**kw)
+    def get_request_url(self,*args,**kw): return self.renderer.get_request_url(self,*args,**kw)
+    def obj2html(self,*args,**kw): return self.renderer.obj2html(self,*args,**kw)
+    def href_to_request(self,*args,**kw): return self.renderer.href_to_request(self,*args,**kw)
+    def row_action_button(self,obj,a,*args,**kw): return self.renderer.row_action_button(obj,self.request,a,*args,**kw)
+        
     def absolute_uri(self,*args,**kw):
         ar = self.spawn(*args,**kw)
         location = ar.renderer.get_request_url(ar)
         return self.request.build_absolute_uri(location)
         
+    def add_system_note(self,owner,subject,body,silent):
+        """
+        System notes are a part of Lino's workflow management system. 
+        A system note is a text message attached to a given 
+        database object instance and propagated through a series of 
+        configurable channels.
+        The text part consists basically of a subject 
+        and a body, both usually generated by the application and edited 
+        by the user in an action's parameters dialog box. 
+        Executing the action will then also trigger the sending of the system note.
+        
+        """
+        #~ logger.info("20121016 add_system_note() '%s'",subject)
+        notes = resolve_app('notes')
+        if notes:
+            notes.add_system_note(self,owner,subject,body)
+        #~ if silent:
+            #~ return
+        recipients = []
+        for addr in settings.LINO.get_system_note_recipients(self,owner,silent):
+            if not '@example.com' in addr:
+                recipients.append(addr)
+        if not len(recipients):
+            return
+        sender = self.get_user().email or settings.SERVER_EMAIL
+        if not sender or '@example.com' in sender:
+            return
+        msg = EmailMessage(subject=subject, 
+            from_email=sender,body=body,to=recipients)
+        msg.send()
+        logger.info("System note '%s' from %s has been sent to %s",subject,sender,recipients)
+
+            
+            
+            
+        
+        
 
 def action(*args,**kw):
     """
@@ -813,6 +1226,5 @@ def action(*args,**kw):
         a.run = fn
         return a
     return decorator
-    
-#~ def action2str(actor,action):
-    #~ return str(actor) + '.' + action.name
+        
+        
