@@ -1,4 +1,4 @@
-#coding: utf-8
+# -*- coding: UTF-8 -*-
 ## Copyright 2008-2012 Luc Saffre
 ## This file is part of the Lino project.
 ## Lino is free software; you can redistribute it and/or modify 
@@ -12,6 +12,10 @@
 ## You should have received a copy of the GNU General Public License
 ## along with Lino; if not, see <http://www.gnu.org/licenses/>.
 
+"""
+Defines classes :class:`Frame` and :class:`FrameHandle`
+"""
+
 
 from django.conf import settings
 from django.db import models
@@ -21,29 +25,24 @@ from django.contrib.contenttypes import generic
 
 import datetime
 
-
-#~ from south.modelsinspector import add_introspection_rules
-#~ add_introspection_rules([], ["^lino\.fields\.LanguageField"])
-#~ add_introspection_rules([], ["^lino\.fields\.PriceField"])
-#~ add_introspection_rules([], ["^lino\.fields\.KnowledgeField"])
-#~ add_introspection_rules([], ["^lino\.fields\.StrengthField"])
-#~ add_introspection_rules([], ["^lino\.fields\.PercentageField"])
-#~ add_introspection_rules([], ["^lino\.fields\.MyDateField"])
-#~ add_introspection_rules([], ["^lino\.fields\.MonthField"])
-#~ add_introspection_rules([], ["^lino\.fields\.QuantityField"])
-#~ add_introspection_rules([], ["^lino\.fields\.HtmlTextField"])
+from decimal import Decimal
 
 
 import logging
 logger = logging.getLogger(__name__)
 
-from lino.tools import full_model_name
-from lino.tools import obj2str
-from lino.tools import resolve_field
+from lino.core.modeltools import full_model_name
+from lino.core.modeltools import obj2str
+from lino.core.modeltools import resolve_field
+from lino.core.modeltools import resolve_model
 
 #~ from lino.utils import choosers
-from lino.utils import choicelists
+#~ from lino.utils import choicelists
 from lino.utils import IncompleteDate, d2iso
+#~ from lino.utils.quantities import Duration
+from lino.utils import quantities 
+from lino.utils import get_class_attr
+
 
 class PasswordField(models.CharField):
     """Stored as plain text in database, but not displayed in user interface."""
@@ -71,6 +70,9 @@ class RichTextField(models.TextField):
     
   
 class PercentageField(models.SmallIntegerField):
+    """
+    Deserves more documentation.
+    """
     def __init__(self, *args, **kw):
         defaults = dict(
             max_length=3,
@@ -80,6 +82,9 @@ class PercentageField(models.SmallIntegerField):
   
 #~ class MonthField(models.CharField):
 class MonthField(models.DateField):
+    """
+    Deserves more documentation.
+    """
     def __init__(self, *args, **kw):
         #~ defaults = dict(
             #~ max_length=10,
@@ -89,6 +94,9 @@ class MonthField(models.DateField):
         models.DateField.__init__(self,*args, **kw)
   
 class PriceField(models.DecimalField):
+    """
+    Deserves more documentation.
+    """
     def __init__(self, *args, **kwargs):
         defaults = dict(
             max_length=10,
@@ -105,49 +113,80 @@ class PriceField(models.DecimalField):
         #~ fld.widget.attrs['style'] = "text-align:right;"
         #~ return fld
         
-class MyDateField(models.DateField):
+#~ class MyDateField(models.DateField):
         
-    def formfield(self, **kwargs):
-        fld = super(MyDateField, self).formfield(**kwargs)
-        # display size is smaller than full size:
-        fld.widget.attrs['size'] = "8"
-        return fld
+    #~ def formfield(self, **kwargs):
+        #~ fld = super(MyDateField, self).formfield(**kwargs)
+        #~ # display size is smaller than full size:
+        #~ fld.widget.attrs['size'] = "8"
+        #~ return fld
         
         
         
-class QuantityField(models.DecimalField):
-    def __init__(self, *args, **kwargs):
-        defaults = dict(
-            max_length=5,
-            max_digits=5,
-            decimal_places=0,
-            )
-        defaults.update(kwargs)
-        super(QuantityField, self).__init__(*args, **defaults)
-        
-    def formfield(self, **kwargs):
-        fld = super(QuantityField, self).formfield(**kwargs)
-        fld.widget.attrs['size'] = "3"
-        fld.widget.attrs['style'] = "text-align:right;"
-        return fld
 
-class FakeField:
+class FakeField(object):
+    """
+    Base class for 
+    :class:`RemoteField`
+    :class:`DisplayField`
+    """
+    choices = None
     primary_key = False
     editable = False
     name = None
     help_text = None
     
     def is_enabled(self,lh):
-        return True
+        """
+        Overridden by mti.EnableChild
+        """
+        return False
+        #~ return True
+        #~ return not self.editable
         
     def has_default(self):
         return False
         
 
+#~ class NullField(FakeField):
+    #~ def __init__(self,name):
+        #~ self.name = name
+    
+class RemoteField(FakeField):
+    """
+    Represents a field on a related object.
+    LayoutHandle instantiates a RemoteField for example when 
+    """
+    #~ primary_key = False
+    #~ editable = False
+    def __init__(self,func,name,fld,**kw):
+        self.func = func
+        self.name = name
+        self.field = fld
+        self.rel = self.field.rel
+        self.verbose_name = fld.verbose_name
+        self.max_length = getattr(fld,'max_length',None)
+        #~ print 20120424, self.name
+
+    def value_from_object(self,obj,ar=None):
+        """
+        Return the value of this field in the specified model instance `obj`.
+        `ar` may be `None`, it's forwarded to the getter method who may 
+        decide to return values depending on it.
+        """
+        m = self.func
+        return m(obj,ar)
+        
+    def __get__(self,instance,owner):
+        if instance is None: return self
+        return self.value_from_object(instance)
+
         
 
 class DisplayField(FakeField):
-    
+    """
+    Deserves more documentation.
+    """
     choices = None
     blank = True
     drop_zone = None
@@ -168,6 +207,9 @@ class DisplayField(FakeField):
 
         
 class HtmlBox(DisplayField):
+    """
+    Deserves more documentation.
+    """
     pass
     
 #~ class QuickAction(DisplayField):
@@ -175,24 +217,67 @@ class HtmlBox(DisplayField):
     
 #~ from django.db.models.fields import Field
 
+VIRTUAL_FIELDS = []
+
+def resolve_virtual_fields():
+    global VIRTUAL_FIELDS
+    for vf in VIRTUAL_FIELDS: 
+        vf.lino_resolve_type()
+    VIRTUAL_FIELDS = None
+      
+
 class VirtualField(FakeField): # (Field):
     """
     Currently subclassed only by :class:`lino.utils.mti.EnableChild`.    
     """
     
     def __init__(self,return_type,get):
-        if isinstance(return_type,basestring):
-            return_type = resolve_field(return_type)
         self.return_type = return_type # a Django Field instance
         self.get = get
-        #~ self.set = set
-        #~ self.name = None
-        #~ Field.__init__(self)
+        VIRTUAL_FIELDS.append(self)
+        """
+        Normal VirtualFields are read-only and not editable.
+        We don't want to require application developers to explicitly 
+        specify `editable=False` in their return_type::
+        
+          @dd.virtualfield(dd.PriceField(_("Total")))
+          def total(self,ar=None):
+              return self.total_excl + self.total_vat
+        """
+            
+    def attach_to_model(self,model,name):
+        self.model = model
+        self.name = name
+        #~ self.return_type.name = name
+        #~ self.return_type.attname = name
+        #~ if issubclass(model,models.Model):
+        #~ self.lino_resolve_type(model,name)
+        model._meta.add_virtual_field(self)
+        #~ logger.info('20120831 VirtualField %s.%s',full_model_name(model),name)
+        
+    def lino_resolve_type(self):
+        """
+        Unlike attach_to_model, this is also called on virtual 
+        fields that are defined on an Actor
+        """
+        #~ logger.info("20120903 lino_resolve_type %s.%s", actor_or_model, name)
+        #~ if self.name is not None:
+            #~ if self.name != name:
+                #~ raise Exception("Tried to re-use %s.%s" % (actor_or_model,name))
+        #~ self.name = name
+        if isinstance(self.return_type,basestring):
+            self.return_type = resolve_field(self.return_type)
+        if isinstance(self.return_type,models.ForeignKey):
+            self.return_type.rel.to = resolve_model(self.return_type.rel.to)
+
+        #~ removed 20120919 self.return_type.editable = self.editable
         for k in ('''to_python choices save_form_data 
           value_to_string verbose_name max_length rel
+          max_digits decimal_places
           blank'''.split()):
-            setattr(self,k,getattr(return_type,k,None))
-            
+            setattr(self,k,getattr(self.return_type,k,None))
+        #~ logger.info('20120831 VirtualField %s on %s',name,actor_or_model)
+      
     def unused_contribute_to_class(self, cls, name):
         ## if defined in abstract base class, called once on each submodel
         if self.name:
@@ -223,19 +308,10 @@ class VirtualField(FakeField): # (Field):
         This special behaviour is needed to implement 
         :class:`lino.utils.mti.EnableChild`.
         """
-        raise NotImplementedError
-        #~ pass
+        raise NotImplementedError("Cannot write %r to field %s" % (value,self))
         
-    def lino_kernel_setup(self,model,name):
-        self.model = model
-        self.name = name
-        #~ self.return_type.name = name
-        #~ self.return_type.attname = name
-        #~ if issubclass(model,models.Model):
-        model._meta.add_virtual_field(self)
-        logger.debug('Found VirtualField %s.%s',full_model_name(model),name)
-        
-    def value_from_object(self,request,obj):
+    #~ def value_from_object(self,request,obj):
+    def value_from_object(self,obj,ar=None):
         """
         Return the value of this field in the specified model instance `obj`.
         `request` may be `None`, it's forwarded to the getter method who may 
@@ -244,29 +320,14 @@ class VirtualField(FakeField): # (Field):
         m = self.get
         #~ assert m.func_code.co_argcount == 2, (self.name, m.func_code.co_varnames)
         #~ print self.field.name
-        return m(obj,request)
+        return m(obj,ar)
         
-class Constant(object):
-    #~ get = None
-    def __init__(self,text_fn):
-        self.text_fn = text_fn
+    def __get__(self,instance,owner):
+        if instance is None: return self
+        return self.value_from_object(instance)
         
-def constant(verbose_name):
-    """
-    Decorator to turn a method into a Constant.
-    """
-    def decorator(fn):
-        def wrapped(*args):
-            #~ return classmethod(fn(*args))
-            return fn(*args)
-        return Constant(wrapped)
-    return decorator
-
-
-class RequestField(VirtualField):
-    def __init__(self,get,*args,**kw):
-        kw.setdefault('max_length',8)
-        VirtualField.__init__(self,DisplayField(*args,**kw),get)
+    def __set__(self,instance,value):
+        return self.set_value_in_object(None,instance,value)
         
 
 def virtualfield(return_type):
@@ -274,11 +335,41 @@ def virtualfield(return_type):
     Decorator to turn a method into a VirtualField.
     """
     def decorator(fn):
-        def wrapped(*args):
-            return fn(*args)
-        return VirtualField(return_type,wrapped)
+        return VirtualField(return_type,fn)
     return decorator
     
+
+
+class Constant(object):
+    """
+    Deserves more documentation.
+    """
+    #~ get = None
+    def __init__(self,text_fn):
+        self.text_fn = text_fn
+        
+#~ def constant(verbose_name=None):
+def constant():
+    """
+    Decorator to turn a method into a :class:`Constant`.
+    """
+    def decorator(fn):
+        #~ def wrapped(*args):
+            #~ return fn(*args)
+        #~ return Constant(wrapped)
+        return Constant(fn)
+    return decorator
+
+
+class RequestField(VirtualField):
+    """
+    Deserves more documentation.
+    """
+    def __init__(self,get,*args,**kw):
+        kw.setdefault('max_length',8)
+        VirtualField.__init__(self,DisplayField(*args,**kw),get)
+        
+
 def displayfield(*args,**kw):
     """
     Decorator shortcut to turn a method into a virtual DisplayField.
@@ -291,9 +382,10 @@ def requestfield(*args,**kw):
     The method to decorate must return either None or a TableRequest object.
     """
     def decorator(fn):
-        def wrapped(*args):
-            return fn(*args)
-        return RequestField(wrapped,*args,**kw)
+        #~ def wrapped(*args):
+            #~ return fn(*args)
+        #~ return RequestField(wrapped,*args,**kw)
+        return RequestField(fn,*args,**kw)
     return decorator
     
         
@@ -310,11 +402,12 @@ class MethodField(VirtualField):
         self.kw = kw
         VirtualField.__init__(self,return_type,get)
         
-    def lino_kernel_setup(self,model,name):
+    def attach_to_model(self,model,name):
         self.get = getattr(model,get)
-        VirtualField.lino_kernel_setup(self,model,name)
+        VirtualField.attach_to_model(self,model,name)
       
-    def value_from_object(self,request,obj):
+    #~ def value_from_object(self,request,obj):
+    def value_from_object(self,obj,ar=None):
         """
         Return the value of this field in the specified model instance `obj`.
         `request` is ignored.
@@ -333,7 +426,7 @@ class MethodField(VirtualField):
         
 
 
-class LinkedForeignKey(generic.GenericForeignKey):
+class unused_LinkedForeignKey(generic.GenericForeignKey):
     """
     Like a GenericForeignKey, but the content type 
     is not stored in another model.
@@ -518,7 +611,7 @@ class GenericForeignKeyIdField(models.PositiveIntegerField):
     for fields that part of a :term:`GFK` and you want 
     Lino to render them using a Combobox.
     
-    Used by :class:`lino.mixins.Owned`.
+    Used by :class:`lino.mixins.Controllable`.
     """
     def __init__(self, type_field, *args, **kw):
         self.type_field = type_field
@@ -527,7 +620,7 @@ class GenericForeignKeyIdField(models.PositiveIntegerField):
 class GenericForeignKey(generic.GenericForeignKey):
     """
     Lino's little extension to Django's GFK.
-    Used by :class:`lino.mixins.Owned`.
+    Used by :class:`lino.mixins.Controllable`.
     """
     def __init__(self, ct_field="content_type", fk_field="object_id", 
           verbose_name=None,help_text=None):
@@ -553,66 +646,55 @@ class GenericForeignKey(generic.GenericForeignKey):
 
 
 
-class ChoiceListField(models.CharField):
-    """
-    A field that stores values from a 
-    :class:`lino.utils.choicelists.ChoiceList`.
+       
+  
+class QuantityField(models.CharField):
+#~ class QuantityField(models.DecimalField):
+#~ class QuantityField(models.Field):
     """
+    A field that accepts both 
+    :class:`lino.utils.quantity.Decimal`,
+    :class:`lino.utils.quantity.Percentage` 
+    and 
+    :class:`lino.utils.quantity.Duration` 
+    values.
     
+    Implemented as a CharField (sorting or filter ranges may not work)
+    
+    QuantityFields are implemented as CharFields and 
+    therefore should *not* be declared `null=True`. 
+    But if `blank=True`, empty strings are converted to `None` 
+    values.    
+  
+    """
     __metaclass__ = models.SubfieldBase
-    
-    #~ choicelist = NotImplementedError
-    
-    def __init__(self,choicelist,*args,**kw):
-        if args:
-            verbose_name = args[0]
-            args = args[1:]
-        else:
-            verbose_name = kw.pop('verbose_name',None)
-        if verbose_name is None:
-            verbose_name = choicelist.label
-        self.choicelist = choicelist
-        defaults = dict(
-            #~ choices=KNOWLEDGE_CHOICES,
-            choices=choicelist.get_choices(),
-            max_length=choicelist.max_length,
-            blank=True,  # null=True,
-            #~ validators=[validate_knowledge],
-            #~ limit_to_choices=True,
-            )
-        defaults.update(kw)
-        #~ models.SmallIntegerField.__init__(self,*args, **defaults)
-        models.CharField.__init__(self,verbose_name,*args, **defaults)
-        
-    def get_internal_type(self):
-        return "CharField"
+    description = _("Quantity (Decimal or Duration)")
+
+    def __init__(self,*args,**kw):
+        kw.setdefault('max_length',6)
+        models.Field.__init__(self,*args,**kw)
+        #~ models.CharField.__init__(self,*args,**kw)
+      
+    #~ def get_internal_type(self):
+        #~ return "CharField"
         
     def to_python(self, value):
-        if isinstance(value,choicelists.BabelChoice):
+        if isinstance(value,Decimal):
             return value
-        value = self.choicelist.to_python(value)
-        if value is None: # see 20110907
-            value = ''
-        return value
-        
-    def get_prep_value(self, value):
         if value:
-            return value.value
-        return '' # see 20110907
-        #~ return None
+            return quantities.parse(value)
+        return None
         
-    def value_to_string(self, obj):
-        value = self._get_val_from_obj(obj)
-        return self.get_prep_value(value)
-        #~ return self.get_db_prep_value(value,connection)
+    #~ def get_db_prep_save(self, value, connection):
+        #~ if value is None:
+            #~ return ''
+        #~ return str(value)
         
-    #~ def save_form_data(self, instance, data):
-        #~ setattr(instance, self.name, data)
+    def get_prep_value(self,value):
+        if value is None:
+            return ''
+        return str(value)
         
-    def get_text_for_value(self,value):
-        return self.choicelist.get_text_for_value(value.value)
-    
-      
 class IncompleteDateField(models.CharField):
     """
     A field that behaves like a DateField, but accepts
@@ -625,8 +707,8 @@ class IncompleteDateField(models.CharField):
         kw.update(max_length=11)
         models.CharField.__init__(self,*args,**kw)
       
-    def get_internal_type(self):
-        return "CharField"
+    #~ def get_internal_type(self):
+        #~ return "CharField"
         
     def to_python(self, value):
         if isinstance(value,IncompleteDate):
@@ -652,5 +734,97 @@ class IncompleteDateField(models.CharField):
         #~ value = self._get_val_from_obj(obj)
         #~ return self.get_prep_value(value)
         
+class DummyField(object):
+    """
+    Deserves more documentation.
+    """
+    def __init__(self,*args,**kw):
+        pass
+    def set_attributes_from_name(self,k):
+        pass
+
+class RecurrenceField(models.CharField):
+    """
+    Deserves more documentation.
+    """
+    #~ __metaclass__ = models.SubfieldBase
+
+    def __init__(self,*args,**kw):
+        kw.setdefault('max_length',200)
+        models.CharField.__init__(self,*args,**kw)
+      
+
+def get_data_elem(model,name):
+    #~ logger.info("20120202 get_data_elem %r,%r",model,name)
+    parts = name.split('__')
+    if len(parts) > 1:
+        # logger.warning("20120406 RemoteField %s in %s",name,self)
+        #~ model = self.model
+        for n in parts:
+            assert model is not None
+            fld = get_data_elem(model,n)
+            if fld is None:
+                # raise Exception("Part %s of %s got None" % (n,model))
+                raise Exception(
+                    "Invalid RemoteField %s.%s (no %s in %s)" % 
+                    (full_model_name(self.model),name,n,full_model_name(model)))
+            if fld.rel:
+                model = fld.rel.to
+            else:
+                model = None
+        def func(obj,ar=None):
+            # logger.info('20120109 %s',name)
+            try:
+                for n in parts:
+                    obj = getattr(obj,n)
+                # logger.info('20120109 %s --> %r', name,obj)
+                return obj
+            except Exception,e:
+                return None
+        return RemoteField(func,name,fld)
+    
+    try:
+        return model._meta.get_field(name)
+    except models.FieldDoesNotExist,e:
+        pass
+        
+    #~ s = name.split('.')
+    #~ if len(s) == 1:
+        #~ mod = import_module(model.__module__)
+        #~ rpt = getattr(mod,name,None)
+    #~ elif len(s) == 2:
+        #~ mod = getattr(settings.LINO.modules,s[0])
+        #~ rpt = getattr(mod,s[1],None)
+    #~ else:
+        #~ raise Exception("Invalid data element name %r" % name)
+    
+    v = get_class_attr(model,name)
+    if v is not None: return v
+    
+    for vf in model._meta.virtual_fields:
+        if vf.name == name:
+            return vf
+
+
+def fields_list(model,field_names):
+    """
+    Return a list with the names of the specified fields, 
+    checking whether each of them exists.
+    
+    Arguments: 
+    `model` is any subclass of `django.db.models.Model`.
+    `field_names` is a single string with a space-separated list of field names.
+    
+    For example if you have a model `MyModel` 
+    with two fields `foo` and `bar`,
+    then ``dd.fields_list(MyModel,"foo bar")`` 
+    will return ``['foo','bar']``
+    and ``dd.fields_list(MyModel,"foo baz")`` will raise an exception.
+    """
+    #~ return tuple([get_field(model,n) for n in field_names.split()])
+    #~ if model.__name__ == 'Company':
+        #~ print 20110929, [get_field(model,n) for n in field_names.split()]
+    #~ return [get_field(model,n).name for n in field_names.split()]
+    return [get_data_elem(model,n).name for n in field_names.split()]
 
 
